#----------------------------------------------------------------
# Copyright (c) 2001 Benjamin Crowell, all rights reserved.
#
# This software is available under two different licenses: 
#  version 2 of the GPL, or
#  the Artistic License. 
# The software is copyrighted, and you must agree to one of
# these licenses in order to have permission to copy it. The full
# text of both licenses is given in the file titled Copying.
#
#----------------------------------------------------------------

package AnswerResponse;

use strict;

use Spotter;
use Expression;
use Parse;
use Eval;
use Rational;
use Units;
use Measurement;

use Math::Complex;
use Math::Trig;

use utf8;

#----------------------------------------------------------------
# answer_response
#----------------------------------------------------------------
# The units are parsed by Units::parse_units(), which doesn't do much
# error checking. 
sub answer_response {
  my ($p,$ans,$units_allowed) = @_;
  my $n_ans = $p->n_ans();
  my @vbl_list = $p->vbl_list();
  my %vbl_hash = $p->vbl_hash();
  my $n_vbl = $#vbl_list;
  my $result = "";
  my $internal_msg = "";

  my $debug = 0;

  if ($debug) {print "entering answer_response"}

  # Parse the units:
  for (my $i=0; $i<=$n_vbl; $i++) {
    my $vbl = $p->get_vbl($vbl_list[$i]);
    my $u = $vbl->units();
    if ($u) {
      $vbl->parsed_units(Units::parse_units(TEXT=>$u));
    }
    else {
      $vbl->parsed_units(Units->new());
    }
  }

  # Try to lex and parse the student's answer. This isn't actually terribly
  # useful, because inside the identical_answer() subroutine we need to
  # parse it all over again using the filter that goes with this canned
  # answer. However, we want to find out right away whether the student
  # made a syntax error.
  if ($debug) {print "answer_response: starting initial error check<p>\n"}
  my $student_expression = Expression->new(EXPR=>$ans,OUTPUT_MODE=>"html",
					VAR_NAMES=>\@vbl_list,UNITS_ALLOWED=>$units_allowed);
  my $parse_error = $student_expression->has_errors(PARSE_IT=>1);
  if ($parse_error) {
    print $student_expression->format_errors();
  }
  if ($debug) {print "answer_response: done with initial error check<p>\n"}

  # In the following, the code for randomly generating values for the variables
  # is duplicated in identical_answer(). Should make this a method of the Problem object.
  if (!$parse_error) {
    my %meas_hash;
    for (my $j=0; $j<=$#vbl_list; $j++) {
      my $sym = $vbl_list[$j];
      my $var = $vbl_hash{$sym};
      my $x = rand_cplx($var->min(),$var->max(),$var->min_imag(),$var->max_imag());
      $meas_hash{$sym} = Measurement->new($x,$var->parsed_units());
      delete($Spotter::standard_cons_hash{$sym});
          	# ...this is obviously stupid, but I have to do it because I've implemented
          	# constants as globals rather than by parameter passing. Note that this won't
          	# have bad consequences later, because we only run with one set of symbols
          	# per invocation of the CGI.
    }
    $student_expression->vars_ref(\%meas_hash);
    #print "aaa".$student_expression->{UNITS_ALLOWED}.".\n";
    my $result = $student_expression->evaluate();
    #print "bbb\n";
    $parse_error = $student_expression->has_errors();
    if ($parse_error) {
      print $student_expression->format_errors();
    }
    if (ref($result) ne "" && ref($result) ne "Math::Complex" && ref($result) ne "Measurement") {
      $parse_error = 1;
      print ref($result)."=";
      print "Illegal variable or use of units. Your expression doesn't result in a number.<br/>";
    }
  }


  # Check whether it matches any of the right or wrong answers we know of.
  # what=1 : Check against the right answer.
  # what=2 : Check if it matches an anticipated wrong answer.
  my $student_answer_is_correct = 0;
  if (!$parse_error) {
    my $done = 0;
    my $fallback_result = "";
    for (my $what=1; $what<=2 && !$done; $what++) {
      for (my $i=0; $i<=$n_ans && !$done; $i++) {
        #print "<br/>what=$what, i=$i ";
        my $canned_ans = $p->get_ans($i);
        my $c = $canned_ans->is_correct();
        if (($what==1 && $c) || ($what==2 && !$c)) {
          my ($match,$msg,$temp_internal_msg) = identical_answer(STUDENT_ANSWER=>$ans,
            VARIABLES=>\%vbl_hash,CANNED_ANSWER=>$canned_ans,
            STUDENT_UNFILTERED_EX=>$student_expression);
          if ($match eq "eq") {
            # It matches:
            $done = 1;
            if ($c) {
              $result = $result .  "<p>Correct</p>\n";
              $student_answer_is_correct = 1;
            }
            else {
              $result = $result .  "<p>Incorrect. ".$canned_ans->response()."</p>\n";
            }
          }
          if ($match eq "er" && $c) {
            $fallback_result =   "<p>$msg</p>\n";
          }
          if ($match eq "un" && $c) {
            $fallback_result =    "<p>Your answer has the wrong units, so either you made a mistake in your algebra or you entered your answer incorrectly.".
                                  " A typical mistake would be to enter a+b/c+d when you really meant (a+b)/(c+d)".
                                  " Scroll down for more information on how to enter answers into Spotter.</p>\n";
          }
          if ($match eq "in" && $c) {
            $fallback_result =    "?"; # internal error
            $internal_msg = $temp_internal_msg;
          }
        }
      }
    }
    if ($result eq "") {$result = $fallback_result;}
    if ($result eq "") {$result = "<p>Incorrect</p>\n"}
    
    if ($result eq "?") {
      Log_file::write_entry(TEXT=>"internal error, $internal_msg");
      $result = "<p>Internal error.</p>\n";
    }
  }
  
  if ($debug) {print "leaving answer_response"}

  return ($result,$student_answer_is_correct);
}

#----------------------------------------------------------------
#           identical_answer
#
#           Test whether the student's answer is the same as the
#           canned one.
#----------------------------------------------------------------
# returns ($what,$message)
#   $what	   eq=matches
#              er=detected error in input
#              va=matches units, not value
#              un=doesn't match units
#              in=internal error
#   $message   relevant for what=2
#   $internal  in case of an internal error, this is what will end up in the log file
#----------------------------------------------------------------
# N_TESTS                how many random values of the variables to test; default is
#                        1 for analytic functions, 10 for nonanalytic
# STUDENT_ANSWER         a string
# CANNED_ANSWER          an Ans object
# VARIABLES              ref to a hash of Vbl objects
# STUDENT_UNFILTERED_EX  an Expression object; this is only used
#                        in order to generate the appropriate error message
#                        when there is an error evaluating (not parsing) 
#                        the student's expression
#----------------------------------------------------------------
sub identical_answer {
  my $n_tests_if_analytic = 1;
  my $n_tests_if_nonanalytic = 10;
  my %args = (
    N_TESTS		=> $n_tests_if_analytic,
    VARIABLES	=> {},
    @_
  );
  my %raw_args = (@_);
  my $n_tests = $args{N_TESTS};
  my $overrode_n_tests_default = (exists $raw_args{N_TESTS});
  my $vars_ref = $args{VARIABLES};
  my $canned_answer = $args{CANNED_ANSWER};
  my $their_string = $args{STUDENT_ANSWER};
  my $their_unfiltered_ex = $args{STUDENT_UNFILTERED_EX};
  my %vars = %$vars_ref;
  my @var_names = keys(%vars);
  my $internal_error = 0;
  my $internal_msg = "";
  my %flags = ();

  my $debug = 0;
  if ($debug) {print "in identical_answer, 1<br/>";}

  # Produce filtered versions of the two expressions:
    my $our_string = $canned_answer->e();
    my $filter = $canned_answer->filter();

    my $x = $filter;
    $x =~ s/\~/\($our_string\)/g;
    $our_string = $x;

    my $x = $filter;
    $x =~ s/\~/\($their_string\)/g;
    $their_string = $x;

    #print "<p>our_string=$our_string, their_string=$their_string, filter=$filter</p>\n";

  # Parse their expression:
    my $their_expression = Expression->new(EXPR=>$their_string,OUTPUT_MODE=>"html",
					VAR_NAMES=>\@var_names);
    my $parse_error = $their_expression->has_errors(PARSE_IT=>1);
    if ($parse_error) {
      $internal_error = 1;
      my $err = $their_expression->format_errors();
      $internal_msg = "identical_answer(): error parsing student's expression $their_string, $err";
      # We already checked the student's input without putting it inside the filter
      # macro. An error at this point means an error in the filter, not in the student's input.
      # Or does it? Conceivably the filter could fail if, e.g., the student's input had
      # the wrong units. OK, but the solution to this is that I should check the units of
      # the student's expression before coming in here.
      # We /don't/ want to give the actual error message, because that reveals information
      # about the filter, and anyway the filter is just an internal thing.
    }

  #print "in identical_answer, 2<br/>";
  # Parse the canned expression:
    my $our_expression;
    if (!$internal_error) {
      $our_expression = Expression->new(EXPR=>$our_string,OUTPUT_MODE=>"html",
					VAR_NAMES=>\@var_names);
      my $parse_error = $our_expression->has_errors(PARSE_IT=>1);
      if ($parse_error) {
        $internal_error = 1;
        $internal_msg = "identical_answer(): error parsing canned expression "
        		.$our_expression->expr()." errors=".$our_expression->format_errors();
      }
    }

  # Test for equality:
  if (      !$overrode_n_tests_default
        and $their_expression->is_nonanalytic() || $our_expression->is_nonanalytic()) {
      $n_tests = $n_tests_if_nonanalytic;
  }
    my $disagreed = 0; # innocent until proven guilty
    if (!$internal_error) {
      for (my $i=1; $i<=$n_tests && !$internal_error && !$disagreed; $i++) {
        my %meas_hash;
        for (my $j=0; $j<=$#var_names; $j++) {
          my $sym = $var_names[$j];
          my $var = $vars{$sym};
          my $x = rand_cplx($var->min(),$var->max(),$var->min_imag(),$var->max_imag());
          $meas_hash{$sym} = Measurement->new($x,$var->parsed_units());
        }
        $our_expression->vars_ref(\%meas_hash);
        $their_expression->vars_ref(\%meas_hash);
        if ($debug) {print "in identical_answer, 99<br/>";}
        my $our_result = Measurement::promote_to_measurement($our_expression->evaluate());
        if ($debug) {print "in identical_answer, 100<br/>";}
        if ($our_expression->has_errors()) {
          $internal_error = 1;
          $internal_msg = "identical_answer(): error parsing, filter=$filter, our_expression=".
          			$our_expression->expr().", our_result=$our_result, err=".
          			$our_expression->format_errors()."=".
          			$our_expression->{HAS_LEX_ERRORS}."=".
          			$our_expression->{HAS_PARSE_ERRORS}."=".
          			$our_expression->{HAS_EVAL_ERRORS}."=".
   					"<br/>";
        }
        else {
          if ($debug) {print "in identical_answer, 199<br/>";}
          my $their_result = Measurement::promote_to_measurement($their_expression->evaluate());
          if ($debug) {print "in identical_answer, 200, their_result=$their_result=, has_errors=".
          			$their_expression->has_errors()."=<br/>";}
          if ($their_expression->has_errors()) {
            $disagreed = 1;
            $flags{EVAL_ERROR} = 1;
          }
          else {
           if (!Measurement::compatible_units($their_result,$our_result,\%Spotter::standard_units)) {
             $disagreed = 1;
             $flags{UNITS_DISAGREE} = 1;
            }
            else {
              my $theirs_converted = Measurement::convert($their_result,$our_result,\%Spotter::standard_units);
              my $our_plain = $our_result;
              if (ref($our_plain) eq "Measurement") {$our_plain = $our_plain->number()}
              my $their_plain = $their_result;
              if (ref($their_plain) eq "Measurement") {$their_plain = $their_plain->number()}

              if ($canned_answer->tol_type() eq "mult" && $our_plain==0) {
                $disagreed = ($their_plain != 0);
              }
              if ($canned_answer->tol_type() eq "mult" && $our_plain!=0 && $their_plain==0) {
		$disagreed = 1;
              }
              if ($canned_answer->tol_type() eq "mult" && $our_plain!=0 && $their_plain!=0) {

                # Figure out the magnitude and argument of theirs/ours:
                  my $ratio = $theirs_converted/$our_result;
                  if ($debug) {print "ratio=$ratio="}
                  my $ratio_with_phase = Crunch::promote_cplx($ratio->number());
                  my $ratio_arg = Math::Complex::arg($ratio_with_phase);
                  my $ratio_mag = Math::Complex::abs($ratio_with_phase);

                # Figure out if ours is real:
                  my $ours_is_real = (ref($our_plain) eq "Complex" && Math::Complex::Im($our_result)==0)
                    || !ref($our_plain);

                # Figure out if theirs is real:
                  my $theirs_is_real = (ref($their_plain) eq "Complex" && Math::Complex::Im($their_plain)==0)
                    || !ref($their_plain);

                if ($debug) {print "ratio_mag=$ratio_mag, ratio_arg=$ratio_arg<br/>\n"}

                # Figure out if magnitudes disagree:
                  my $epsilon = $canned_answer->tol();
                  if ($ratio_arg>1) {$ratio_mag = 100000.}
                  if ($ratio_mag<1.) {$ratio_mag=1/$ratio_mag}
                  if ($ratio_mag>1+$epsilon) {
                    $disagreed = 1;
                  }
                 if ($debug) {print "ratio_mag=$ratio_mag, epsilon=$epsilon, disagreed=$disagreed<br/>\n"}

                # Even if the magnitudes agreed, we may have to do more:
                  if (!$disagreed) {
                    if ($ours_is_real && !$theirs_is_real) {
                      $disagreed = ($ratio_arg>$epsilon) || ($ratio_arg>0.00001);
                    }
                    if (!$ours_is_real && !$theirs_is_real) {
                      $disagreed = ($ratio_arg>$epsilon);
                    }
                  }

              }

              if ($canned_answer->tol_type() eq "add") {
                my $diff = $theirs_converted-$our_result;
                if (abs($diff->number())>$canned_answer->tol()) {
                  $disagreed = 1;
                }
              }

              if ($disagreed) {
               $flags{NUMERICAL_DISAGREEMENT} = 1;
              }

            }
          }
        }
      }
   }
  
  if ($internal_error) {Log_file::write_entry(TEXT=>$internal_msg); return ("in","",$internal_msg)}
  if (!$disagreed) {return ("eq","","")}
  my $msg = "";
  # Should improve the following so it passes on actual messages from the parser:
  if (exists($flags{EVAL_ERROR})) {return("er"," error evaluating","")}
  if (exists($flags{NUMERICAL_DISAGREEMENT})) {return("va","","")}
  if (exists($flags{UNITS_DISAGREE})) {return("un","","")}
}

sub rand_cplx {
  my ($min,$max,$min_imag,$max_imag) = @_;
  return rand_range($min,$max)+i*rand_range($min_imag,$max_imag);
}
sub rand_range {
  my ($min,$max) = @_;
  return rand(1)*($max-$min)+$min;
}

1;
