#!/usr/bin/perl

#----------------------------------------------------------------
# Copyright (c) 2001 Benjamin Crowell, all rights reserved.
#
# This software is available under two different licenses: 
#  version 2 of the GPL, or
#  the Artistic License. 
# The software is copyrighted, and you must agree to one of
# these licenses in order to have permission to copy it. The full
# text of both licenses is given in the file titled Copying.
#----------------------------------------------------------------

package Rational;
use strict vars;

# The Rational class stores a rational number, i.e. the quotient of
# two integers. The operators +-*/ and a few more are overloaded,
# but cmp and <=> isn't (see below).
#
# I wanted to roll my own as a way to learn OO programming in Perl, but
# there are much better implementations out there, such as the Fraction
# package, available from CPAN.
#
# Currently, all methods assume arguments are supplied in simplified form,
# but external code shouldn't depend on the assumption that rationals are
# always simplified. In other words, don't use numer and denom directly,
# except when creating a new object.

use overload '+' => \&add,'-' => \&subtract,'*' => \&mult,'/' => \&div,
		 '""' => \&stringify, 'abs'=>absval, '<=>'=>compare;

# new(numer,denom), or new(int) or new(float)
# For new(float), an error can occur, in which case it returns a null string.
sub new {
  my $classname = shift;
  my $self = {};
  bless($self,$classname);
  my $error = 0;
  if ($#_ == 1) { # 2 args
    $self->numer(shift);
    $self->denom(shift);
  }
  else {
    my $x = shift;
    if (int($x)==$x) {
      $self->numer($x);
      $self->denom(1);
    }
    else {
      my $did = 0;
      for (my $i = -10; $i<=10 && !$did; $i++) {
        for (my $j = 1; $j<=10 && !$did; $j++) {
          if (abs($x-$i/$j)<0.0001) {
            $self->numer($i);
            $self->denom($j);
            $did = 1;
          }
        }
      }
      $error = !$did;
    }
  }
  if (!$error) {
    $self->simplify;
    return $self;
  }
  else {
    return "";
  }
}

sub numer {
  my $self = shift;
  if (@_) {$self->{NUMER} = shift;}
  return $self->{NUMER};
}

sub denom {
  my $self = shift;
  if (@_) {$self->{DENOM} = shift;}
  return $self->{DENOM};
}

sub float {
  my $self = shift;
  return $self->numer/$self->denom;
}

sub stringify {
  my $self = shift;
  if ($self->denom != 1) {
    return $self->numer . "/" . $self->denom;  
  }
  else {
    return $self->numer;  
  }
}

sub absval {
  my ($a) = @_;
  return Rational->new(abs($a->numer),abs($a->denom));
}

sub add {
  my ($a,$b) = @_;
  my $c = Rational->new($a->numer*$b->denom+$a->denom*$b->numer,$a->denom*$b->denom);
  $c->simplify;
  return $c;
}

sub subtract {
  my ($a,$b) = @_;
  my $c = Rational->new($a->numer*$b->denom-$a->denom*$b->numer,
  			$a->denom*$b->denom);
  $c->simplify;
  return $c;
}

sub is_integer {
  my $self = shift;
  $self->simplify;
  return ($self->denom == 1);  
}

# doesn't work yet:
sub compare {
  my ($a,$b) = @_;
  my $c = $a-$b;
  #print "in compare, " . $a . "," . $b. "," . $c . "\n";
  return sign $c;
}

sub sign {
  my $self = shift;
  return ($self->numer) <=> 0; # assumes already simplified
}



sub mult {
  my ($a,$b) = @_;
  my $c = Rational->new($a->numer*$b->numer,$a->denom*$b->denom);
  $c->simplify;
  return $c;
}

sub div {
  my ($a,$b) = @_;
  my $c = Rational->new($a->numer*$b->denom,$a->denom*$b->numer);
  $c->simplify;
  return $c;
}

sub simplify {
  my $self = shift;
  if ($self->numer==0) {
    $self->denom(1);
    return;
  }
  if ($self->denom<0) {
    $self->numer(-$self->numer);
    $self->denom(-$self->denom);
  }
  # mod operator messes up on negatives, so temporarily force positive:
  my $s = 1;
  if ($self->numer<0) {
    $s = -1;
    $self->numer(-$self->numer);
  }
  for (my $i=2; $i<=$self->numer; $i++) {
    while (($self->numer % $i)==0 && ($self->denom % $i)==0) {
      my ($n,$d) = ($self->numer,$self->denom);
      $self->numer($n/$i);
      $self->denom($d/$i);
    }
  }
  $self->numer($s*$self->numer);
}

#==================================================================
#==================================================================
#==================================================================


1;
