#!/usr/bin/perl

#----------------------------------------------------------------
# Copyright (c) 2001 Benjamin Crowell, all rights reserved.
#
# This software is available under two different licenses: 
#  version 2 of the GPL, or
#  the Artistic License. 
# The software is copyrighted, and you must agree to one of
# these licenses in order to have permission to copy it. The full
# text of both licenses is given in the file titled Copying.
#
#----------------------------------------------------------------

use strict;

use Spotter;
use Login;
use FileTree;
use Expression;
use Parse;
use Eval;
use Rational;
use Units;
use Measurement;
use Journal;
use WorkFile;
use BulletinBoard;
use Email;
use AnswerResponse;
use SpotterText;
use SpotterHTMLUtil;

use Math::Complex;
use Math::Trig;
use Message;
use Getopt::Std;
use XML::Parser;
use CGI;
use Digest::SHA1;

use utf8;

#$| = 1; # Set output to flush directly (for troubleshooting)

#----------------------------------------------------------------
# Initialization
#----------------------------------------------------------------
$SpotterHTMLUtil::cgi = new CGI;
Url::decode_pars();

my $tree = FileTree->new(DATA_DIR=>"spotter/",CLASS=>Url::par("class"));

#----------- subroutine for checking passwords
sub my_password_checker {
       my $login = shift;
       if (!$tree->class()) {$login->login_error("No such class."); return 0}
       my ($err,$password_hash) = $tree->get_par_from_file($tree->student_info_file_name($login->username()),"password");
       if ($err) {$login->login_error("No such user, $err."); return 0}
       if (!$login->check_password($password_hash)) {$login->login_error("Incorrect password.");return 0}
       return 1;
}

our $basic_file_name = "spotter";
if (Url::par_set("file")) {
  $basic_file_name = Url::par("file");
}
our $xmlfile = $basic_file_name.".xml";
Log_file::set_name($basic_file_name,"log");# This may be changed by a <log_file> tag.


my $current_problem;
my $printed_any_problems_yet = 0;

if (Url::par_set("debug")) {Html::activate_debugging_output()}

#------------ initialize some login information --------------
our $login = Login->new(
   CGI=>$SpotterHTMLUtil::cgi,
   COOKIE_NAME=>'spotter_login',
   HASH_STRING=>'spotter',
   PASSWORD_CHECKER=>\&my_password_checker
);

#----------------------------------------------------------------
# Contents of page.
#----------------------------------------------------------------

#----------------------------------------------------------------
# Before we write the headers, we need to check whether to set or
# delete a cookie.
#----------------------------------------------------------------
  my $cookie_list = [];
  my $delete_cookie = Url::par_is("login","form"); # can also get set in a couple of other places below
  if (Url::par_is("login","set_cookie")) {
    # The cookie is being set, but isn't set yet, so the normal
    # thing where we read the cookie to check if we're logged in doesn't work.
    # Calling cookie_value() below has the side-effect of
    # setting login->auth.
    my $password = $SpotterHTMLUtil::cgi->param('password');
    $login->username($SpotterHTMLUtil::cgi->param('username'));
    $login->date($SpotterHTMLUtil::cgi->param('date'));
    my $cookie_value = $login->cookie_value($password);
    $login->logged_in(my_password_checker($login));
    if (!$login->logged_in() && $password =~ m/^\@(\d+)$/) {
      # Fullerton College IDs have an extra @ on the front. Don't choke if they type in the @ sign.
      $password = $1;
      $login->login_error('');
      $cookie_value = $login->cookie_value($password);
      $login->logged_in(my_password_checker($login));
    }
    if ($login->logged_in()) {
      if ($SpotterHTMLUtil::cgi->param('email') ne '') {
        $tree->set_par_in_file($tree->student_info_file_name($login->username()),'email',$SpotterHTMLUtil::cgi->param('email'));
        $tree->set_par_in_file($tree->student_info_file_name($login->username()),'emailpublic',
                                                             ($SpotterHTMLUtil::cgi->param('emailpublic') eq 'public'));
      }
      if ($SpotterHTMLUtil::cgi->param('newpassword1') ne '') {
        my ($p1,$p2) = ($SpotterHTMLUtil::cgi->param('newpassword1'),$SpotterHTMLUtil::cgi->param('newpassword2'));
        if ($p1 ne $p2) {
          $login->login_error("You didn't type the same password twice. Please use the back button in your browser and try again.");
          $login->logged_in(0);
        }
        if ($p1 eq '' && $tree->get_student_par($login->username(),'state') eq 'notactivated') {
          $login->login_error("You didn't enter a password. Please use the back button in your browser and try again.");
          $login->logged_in(0);
        }
        if ($login->logged_in() && $p1 ne '') {
          $password = $p1;
          $cookie_value = $login->cookie_value($password);
          $tree->set_par_in_file($tree->student_info_file_name($login->username()),'state','normal');
          $tree->set_par_in_file($tree->student_info_file_name($login->username()),'password',
                              Login::hash($login->hash_string().$password));
        }
      }
      my $cookie = $SpotterHTMLUtil::cgi->cookie(-name=>$login->cookie_name(),-value=>$cookie_value);
      # By default, cookie expires when the browser exits.
      my @foo = ($cookie);
      $cookie_list = \@foo;
    }
    else {
      $delete_cookie = 1;
    }
  }
  if (Url::par_is("login","log_out")) {
    $delete_cookie = 1;
    $login->logged_in(0);
  }
  if ($delete_cookie) {
      my $cookie_deletion = $SpotterHTMLUtil::cgi->cookie(-name=>$login->cookie_name(),-value=>'',-expires=>'-1y');
      my @foo = ($cookie_deletion);
      $cookie_list = \@foo;
    }

#----------------------------------------------------------------
# Headers.
#----------------------------------------------------------------
SpotterHTMLUtil::PrintHTTPHeader($cookie_list);
SpotterHTMLUtil::PrintHeaderHTML();

#print "<h1>Spotter</h1>\n";
print SpotterHTMLUtil::BannerHTML($tree);

#Url::report_pars();
SpotterHTMLUtil::debugging_output("The log file is ".Log_file::get_name());
SpotterHTMLUtil::debugging_output("The xml file is ".$xmlfile);

SpotterHTMLUtil::debugging_output("Logged in: ".$login->logged_in());
SpotterHTMLUtil::debugging_output("Username: ".$login->username());

SpotterHTMLUtil::debugging_output("class_err=".$tree->class_err());
SpotterHTMLUtil::debugging_output("class_description=".$tree->class_description());

my $fatal_error = "";
my $parsed_any;

#----------------------------------------------------------------
# Body.
#----------------------------------------------------------------

  if ($tree->class_description()) {print $tree->class_description()."<br>\n"}

  my $journals = $tree->journals();
  my $have_journals = defined $journals;
  my (%journals,@journals_list);
  if ($have_journals) {
    my ($a,$b) = ($journals->[0],$journals->[1]);
    %journals = %$a;
    @journals_list = @$b;
  }

  my $have_inbox = 0;
  my $have_unread_messages = 0;

  if ($login->logged_in()) {
    $have_inbox = $tree->student_inbox_exists($login->username());
    if ($have_inbox) {
      $have_unread_messages = BulletinBoard::has_unread_messages($tree,$login->username());
    }
  }

  if ($have_unread_messages) {
    my @message_keys = BulletinBoard::unread_message_keys($tree,$login->username()); # sort by date
    @message_keys = sort @message_keys;
    foreach my $key(@message_keys) {
      my $msg = BulletinBoard::get_message($tree,$login->username(),$key);
      if ($msg->[0] eq '') {
        print BulletinBoard::html_format_message($msg);
        my $date = BulletinBoard::current_date_for_message_key();
        BulletinBoard::mark_message_read($tree,$login->username(),$key,$date);
      }
    }
  }

  if (!Url::par_is("login","form")) {
    if ($login->logged_in()) {
      my $have_workfile = -e ($tree->student_work_file_name($login->username()));
      print
               "<b>".$tree->get_real_name($login->username(),"firstlast")."</b> logged in "
              ." | <a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'account',DELETE=>'(login|journal|send_to)')."\">account</a>"
              ." | <a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'grades', DELETE=>'(login|journal|send_to)')."\">grades</a>"
              ." | <a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'email',  DELETE=>'(login|journal|send_to)')."\">e-mail</a>";
      if ($have_journals) {
          print
               " | <a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'edit',DELETE=>'(login|journal|send_to)')."\">edit</a>";
        }
      if ($have_workfile) {
          print
               " | <a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'answers',DELETE=>'(login|journal|send_to)')."\">answers</a>";
      }
      print
               " | <a href=\"".Url::link(REPLACE=>'login',REPLACE_WITH=>'log_out',
                                            REPLACE2=>'what',REPLACE_WITH2=>'check',DELETE_ALL=>1)."\">log out</a><br>\n";
    }
    else {
      print "not logged in | <a href=\"".Url::link(REPLACE=>'login',REPLACE_WITH=>'form')."\">log in</a><br>\n";
      print "(You can check your answers to problems without being logged in.)<p>\n";
    }
  }

  if ($have_journals && Url::par_is("what","edit")) {
    my $first_one = 1;
    print "<p><b>edit</b> ";
    foreach my $j(@journals_list) {
      if (!$first_one) {print " | "}
      $first_one = 0;
      print
               "<a href=\"".Url::link(REPLACE=>'what',REPLACE_WITH=>'edit',DELETE=>'login',
                                            REPLACE2=>'journal',REPLACE_WITH2=>$j)."\">".$journals{$j}."</a>";
    }
    print "<p>\n";
  }

  if (Url::par_is("login","form")) {
    do_login_form();
  }
  if (Url::par_is("login","set_cookie")) {
    SpotterHTMLUtil::debugging_output("username=".$SpotterHTMLUtil::cgi->param('username'));
    SpotterHTMLUtil::debugging_output("password=".$SpotterHTMLUtil::cgi->param('password'));
  }

  if (Url::par_set("login") && !Url::par_is("login","form") && $login->login_error()) {
    print "<p><b>Error: ".$login->login_error()."</b></p>\n";
  }

  if (Url::par_set("login") && !Url::par_is("login","log_out") && $tree->class_err()) {
    print "<p><b>Error: ".$tree->class_err()."</b></p>\n";
  }


  # Set up XML parser:
  my $parser = new XML::Parser(ErrorContext => 2);

  $parser->setHandlers(
        Char=>\&char_handler,
        Start=>\&start_handler,
        End=>\&end_handler);

  $parsed_any = 0;

#----------------------------------------------------------
# Do real stuff.
#----------------------------------------------------------
  if (!$fatal_error && !Url::par_is("login","form")) {
    if (Url::par_is("what","check")) {
      # Seed the random number generator. We don't really want the numbers to
      # be random. We want the same random numbers to be used every time a particular
      # answer is checked. That way the comparison of input to stored answers is
      # deterministic. Since the CGI only ever checks one problem, this is guaranteed
      # to be deterministic as long as we always seed the random number generator
      # the same way. I was tempted to re-seed the generator every time through
      # answer_response(), but Perl doesn't allow srand to be called more than once,
      # and it doesn't matter because we only check one problem per run.
      srand(1);
      if (!(-e $xmlfile)) {$fatal_error = "file '$xmlfile' does not exist"}
      if (!(-r $xmlfile)) {$fatal_error = "unable to read file '$xmlfile'"} # This doesn't seem to work; see below.
      if (!$fatal_error && !(-e Log_file::get_name())) {$fatal_error = "unable to open log file '".Log_file::get_name()."'"}
      if (!$fatal_error && !(-w Log_file::get_name())) {$fatal_error = "unable to write to log file '".Log_file::get_name()."'"}
      $parser->parsefile($xmlfile);
      if (!$parsed_any) {$fatal_error = "unable to read file '$xmlfile'"}
    }
    if (Url::par_is("what","emailpassword")) {
      my $username = $SpotterHTMLUtil::cgi->param('username');
      my $id_given = $SpotterHTMLUtil::cgi->param('id');
      $id_given =~ s/^\@//; # get rid of initial @ sign in Fullerton College student ID
      my $id = $tree->get_student_par($username,'id');
      my $email = $tree->get_student_par($username,'email');
      my $bogus = '';
      if (! Email::syntactically_valid($email)) {$bogus = "You have not given a valid e-mail address, so you can't reset your password.";}
      if ($id_given ne $id) {$bogus = "Incorrect student ID.";}
      if ($bogus eq '') {
        srand();
        my $key = int(rand 1000).int(rand 1000).int(rand 1000);
        $tree->set_student_par($username,'newpasswordkey',$key);
        my $password = Login::hash($login->hash_string().$id);
        $tree->set_student_par($username,'password',$password);
        Email::send_email(TO=>$email,SUBJECT=>'Spotter password',
             BODY=>("To reset your password, please go the following web address:\n".
                 Url::link(REPLACE=>'what',REPLACE_WITH=>'resetpassword',REPLACE2=>'username',REPLACE_WITH2=>$username,
                            REPLACE3=>'key',REPLACE_WITH3=>$key,RELATIVE=>0)));
        print "An e-mail has been sent to you with information on how to set a new password.<p>\n";
      }
      else { # We also end up here if the username is null or invalid.
        print "$bogus<p>\n";
      }
    }
    if (Url::par_is("what","resetpassword")) {
      my $username = Url::par('username');
      my $key = Url::par('key');
      my $key2 = $tree->get_student_par($username,'newpasswordkey');
      if ($key eq $key2 && $key2 ne '') {
        $tree->set_student_par($username,'state','notactivated');
        $tree->set_student_par($username,'newpasswordkey','');
        print "Your account has been inactivated, and you can now reactivate it by typing in your student ID and choosing ";
        print 'a new password. <a href="'.Url::link(REPLACE=>'login',REPLACE_WITH=>'form',REPLACE2=>'what',REPLACE_WITH2=>'check',
                                           DELETE=>'key').'">';
        print "Click here</a> to reactivate your account.<p>\n";
      }
      else {
        print "Error: invalid key.<p>\n";
      }
    }

    if (Url::par_is("what","account")) {
      do_account();
    }

    if (Url::par_is("what","email")) {
      my $bogus = '';
      my $own_email = '';
      my $username = '';
      if (!$login->logged_in()) {
        $bogus = "You must be logged in to send e-mail through Spotter.";
      }
      else {
        $username = $login->username();
        $own_email = $tree->get_student_par($username,'email');
        if ( ! Email::syntactically_valid($own_email)) {
          $bogus = "You can't send e-mail through Spotter because you haven't supplied a valid "
                  ."address yourself. To set your own address, click on the account link above.";
        }
      }
      if ($bogus eq '') {
        do_email($username,$own_email);
      }
      else {
        print "<p>$bogus</p>\n";
      }
    }

    if (Url::par_is("what","grades")) {
      if ($login->logged_in()) {
        do_grades();
      }
      else {
        print "You must be logged in to check your grade.<p>\n";
      }
    }

    if (Url::par_is("what","answers")) {
      if ($login->logged_in()) {
        do_answers();
      }
      else {
        print "You must be logged in to look at your answers.<p>\n";
      }
    }

    if (Url::par_is("what","edit")) {
      my $which_journal = Url::par("journal");
      if ($which_journal ne '') {
        if ($login->logged_in()) {
          do_edit_journal($which_journal);
        }
        else {
          print "You must be logged in to edit.<p>\n";
        }
      }
    }

    if (Url::par_is("what","viewold")) {
      my $which_journal = Url::par("journal");
      if ($which_journal ne '') {
        if ($login->logged_in()) {
          do_view_old($which_journal);
        }
      }
    }

  }


#---------------------------------------------------------
# Footer.
#---------------------------------------------------------

if ($fatal_error) {
  print "<p>Error: $fatal_error</p>\n";
}

print SpotterHTMLUtil::FooterHTML($tree);

# Automatically adds one to month, so Jan=1, and, if year is less than
# 1900, adds 1900 to it. This should ensure that it works in both Perl 5
# and Perl 6.
sub current_date {
    my $what = shift; #=day, mon, year, ...
    my @tm = localtime;
    if ($what eq "sec") {return $tm[0]}
    if ($what eq "min") {return $tm[1]}
    if ($what eq "hour") {return $tm[2]}
    if ($what eq "day") {return $tm[3]}
    if ($what eq "year") {my $y = $tm[5]; if ($y<1900) {$y=$y+1900} return $y}
    if ($what eq "month") {return ($tm[4])+1}
}

sub current_date_string() {
    return sprintf "%04d-%02d-%02d %02d:%02d:%02d", current_date("year"), current_date("month") ,
    current_date("day"),current_date("hour"),current_date("min"),current_date("sec");
}

sub current_date_string_no_time() {
    return sprintf "%04d-%02d-%02d", current_date("year"), current_date("month") ,
    current_date("day");
}



sub do_edit_journal {
  my $journal = shift;
  my $username = $login->username();
  my ($text,$is_locked) = $tree->read_journal($username,$journal);
  my $edited_version = $SpotterHTMLUtil::cgi->param('journalText');
  if ($edited_version || $SpotterHTMLUtil::cgi->param('submitJournalButton')) { # The second clause allows users to submit an empty journal.
    if (!$is_locked) {
      $text = $edited_version;
      $text =~ s/\015?\012/\n/g;
      # Break up long lines:
      for (my $i=1; $i<=30; $i++) {
        $text =~ s/\n([^=*][^\n]{80,}) (\w[^\n]{80,})/\n$1\n$2/g;
      }
      $text =~ s/\n/\012/g;
      if (!($text=~m/\n$/)) {$text = "$text\n"}
      $tree->write_journal($username,$journal,$text);
    }
  }
  my $cooked_text = Journal::format_journal_as_html($text);
  print "<p/>If you scroll down, first you'll see your current version of your text with all the formatting, and then below that you'll ";
  print "see a window in which you can edit your text. To make a paragraph break, put in a blank line between the paragraphs. ";
  print "To make a section heading, put the heading on a line by itself, with an equals sign, =, at the beginning of the line. ";
  print "Subsection headings are made with a ==, and subsubsections with a ===. ";
  print "To make a table of data, put a * at the beginning of each line.<p/>\n";
  print "Your changes will not be saved until you click on the Save button! To avoid losing changes by mistake, you should make ";
  print "a habit of saving your text very often as you work on it.<p/>\n";
  print "<h2>Last Saved Version</h2>".$cooked_text."<p/>\n";
  print "<h2>Edit</h2>\n";
  if ($is_locked) {
    print "This is your final version, and it can no longer be edited.<p/>\n";
  }
  else {
    print '<form method="POST" action="'.Url::link().'">'."\n";
    print '<textarea name="journalText" rows="30" cols="85">';
    print $text;
    print "</textarea><br/>\n";
    print '<input type="submit" name="submitJournalButton" value="Save">'."\n";
    print "</form>\n";
  }
  my $diffs_dir = $tree->diffs_directory();
  my $journal_diffs_dir = "$diffs_dir/$username/$journal";
  if (-e $diffs_dir) {
    my @diffs = sort <$journal_diffs_dir/*>;
    if (@diffs) {
      my $n = @diffs;
      $n--;
      my $url = Url::link(REPLACE=>'what',REPLACE_WITH=>'viewold',DELETE=>'login',
													REPLACE2=>'journal',REPLACE_WITH2=>$journal);
      print "<h3>Old Versions</h3>\nYou have $n old versions you can go back and look at. To view one, enter a number from 1 to $n\n<br/>\n";
      print '<form method="POST" action="'.$url.'">'."\n";
      print '<input type="text" name="version">'."\n";
      print '<input type="submit" name="oldJournalButton" value="View">'."\n";
      print "</form>\n";
      print "<p><b>If you have edited your text, make sure to save it before doing this!</b></p>\n";
    }
  }
}

sub do_view_old {
  my $journal = shift;
  my $version = $SpotterHTMLUtil::cgi->param('version');
  my $username = $login->username();
  print "<h2>Viewing old version $version of $journal</h2>\n";
  my $diffs_dir = $tree->diffs_directory();
  my $journal_diffs_dir = "$diffs_dir/$username/$journal";
  my @diffs = sort <$journal_diffs_dir/*>;
  # create empty file:
  my $rebuild = "$journal_diffs_dir/rebuild";
  open(FILE,">$rebuild");
  close(FILE);
  for (my $i=0; $i<=$version-1; $i++) {
    system("patch -s $rebuild $diffs[$i]");
  }
  unlink "$rebuild.orig";
  local $/;
  open(FILE,"<$rebuild");
  my $text = <FILE>;
  close FILE;
  print "<pre>$text</pre>";
}

sub do_answers {
  my ($err,$answers) = WorkFile::list_all_correct_answers_for_one_student($tree,$login->username());
  if ($err eq '') {
    my @answers = @$answers;
    print "<p>The following is a list of the correct answers that have been recorded for you. If a correct answer ";
    print "is missing from this list, it may be because you weren't logged in when you entered the answer. Even if ";
    print "your correct answers shows up here, that doesn't necessarily mean it was on time. Note that all the times shown below ";
    print "are for the time zone of the server (central time for lightandmatter.com). If you got some parts of a problem ";
    print "right but not others, only the ones you got right are listed here.</p>\n";
    foreach my $answer(@answers) {
      print "$answer<br/>\n";
    }
  }
  else {
    print "<p>Error: $err</p>\n";
  }
}

sub do_grades {
  local $/; # slurp the whole file
  my $err = 0;
  my $username = $login->username();
  my $filename = $tree->grade_report_file_name($username);
  open(REPORT,"<$filename") or $err=1;
  if ($err) {print "Error opening grade report.<p>\n"; return}
  print <REPORT>;
  close REPORT;
}

sub do_email {
      my $username = shift;
      my $own_email = shift;

      if (Url::par_set("send_to")) {
        my $own_name = $tree->get_real_name($username,"firstlast");
        my $subject1 = "Email via Spotter";
        if ($tree->class_description()) {$subject1 = $tree->class_description()}
        my $subject2 = $SpotterHTMLUtil::cgi->param('emailSubject');
        my $link = Url::link();
        my $body = $SpotterHTMLUtil::cgi->param('emailBody');
        #print "calling, link=$link,body=$body,sub1=$subject1,sub2=$subject2<p>\n";
        Email::send_email_from_student($username,$own_email,$own_name,Url::par("send_to"),$link,$body,
               $subject1,$subject2);
        return;
      }

      my @roster = $tree->get_roster();
      print "<p><b>E-mail addresses</b><p>\n";
      print "Important privacy information: People's e-mail addresses only appear here if they want them to be available to ";
      print "other people in the class; this can be controlled from the account settings page. Please do not give these e-mail ";
      print "addresses to anyone outside the class.<p>\n";
      print "<table>\n";
      my $instructor_emails = $tree->instructor_emails();
      my @instructors = keys %$instructor_emails;
      if (@instructors) {
        my $form = "instructors";
        if (@instructors==1) {$form = "instructor"}
        print "<tr><td><i>$form</i></td></tr>\n";
      }
      foreach my $who(@instructors) {
        my $email = $instructor_emails->{$who};
        print "<tr>";
        print "<td>$who</td>\n<td>";
        if ($email ne '') {
          print link_to_send_email($email);
        }
        else {
          print '---';
        }
        print "</td></tr>\n";
        
      }
      print "<tr><td><i>students</i></td></tr>\n";
      foreach my $who(@roster) {
        print "<tr>";
        print "<td>".$tree->get_real_name($who,"lastfirst")."</td>\n<td>";
        if (Email::syntactically_valid($tree->get_student_par($who,"email")) && $tree->get_student_par($who,"emailpublic")) {
          my $email = $tree->get_student_par($who,"email");
          print link_to_send_email($email);
        }
        else {
          if (!Email::syntactically_valid($tree->get_student_par($who,"email"))) {
            print '---'; # no (syntactically valid) address given
          }
          else {
            print '(not public)'; # address given, but not public
          }
        }
        print "</td></tr>\n";
      }
      print "</table>\n";
}

sub link_to_send_email {
  my $address = shift;
  return " <a href=\""
     .Url::link(REPLACE=>'what',REPLACE_WITH=>'email',  DELETE=>'(login|journal)',REPLACE2=>'send_to',REPLACE_WITH2=>$address)
     ."\">Send e-mail.</a>";

}

sub do_account {
      print "<p><b>Your Account</b><p>\n";
      print '<form method="POST" action="'.Url::link(REPLACE=>'login',REPLACE_WITH=>'set_cookie',
                                                     REPLACE2=>'what',REPLACE_WITH2=>'check').'">';

      my $email = $tree->get_student_par($login->username(),'email');
      my $emailpublic = $tree->get_student_par($login->username(),'emailpublic');
      print "<p>E-mail address:<br>";
      print '<input type="text" name="email" size="50" maxlength="50" value="'.$email.'"><br>';
      print '<input type="checkbox" name="emailpublic" ';
      if ($emailpublic) {print 'checked'}
      print ' value="public"> Check this box if you want other students in ';
      print 'the class to have access to this e-mail address.<br>';

      print "<p>If you want to change your password, enter the new one twice below. If you don't want to change your password, don't ";
      print "type in these boxes.<br>";
      print '<table><tr><td>New password:</td><td><input type="password" name="newpassword1" size="20" maxlength="20"></td></tr>';
      print '<tr><td>New password again:</td>'
                                        .'<td><input type="password" name="newpassword2" size="20" maxlength="20"></td></tr></table>';

      print '<p><i>The current settings on your account are given above. To change them, edit the form and then enter ';
      print "your password at the bottom of the form and press the Change Settings button.</i>\n";
      print '<p>Password:  <input type="password" name="password" size="20" maxlength="20">';
      print "(If you're changing your password, enter your old one here.)<br>";
      print '  <input type="submit" value="Change Settings">';
      print "</form>\n";
}

sub do_login_form {
  my $step = 1;
  my $username = '';
  my $disabled = 0;
  my $state = '';
  if (Url::par_set("class") && !$tree->class_err()) {
    $step = 2;
    $username = Url::par('username');
    $disabled = ($tree->get_student_par($username,'disabled'));
    $state = ($tree->get_student_par($username,'state'));
    if ($username) {$step=3}
  }
  if ($disabled) {
    print "Your account has been disabled.<p>\n";
  }
  else {
    if ($step==1) { # set class
      print "To log in, start from the link provided on your instructor's web page.<p>\n";
      # has to have &class=bcrowell/f2002/205 or whatever in the link
    }
    if ($step==2) { # set username
      my @roster = $tree->get_roster();
      print "<p><b>Click on your name below:</b><br>\n";
      foreach my $who(@roster) {
        print '<a href="'.Url::link(REPLACE=>'username',REPLACE_WITH=>$who).'">';
        print $tree->get_real_name($who,"lastfirst");
        print "</a><br>\n";
      }
    }
    if ($step==3) { # enter password, and, if necessary, activate account
      print "<b>".$tree->get_real_name($username,"firstlast")."</b><br>\n";
      my $date = current_date_string();
      print '<form method="POST" action="'.Url::link(REPLACE=>'login',REPLACE_WITH=>'set_cookie').'">';
      print '  <input type="hidden" name="username" value="'.$username.'">';
      print '  <input type="hidden" name="date" value="'.current_date_string().'">';
      if ($state eq 'normal') {print ' Password:'} else {print ' Student ID:'} # initially, password is student ID
      print '  <input type="password" name="password" size="20" maxlength="20"><br>';
      if ($state ne 'normal') {
        print '<p><i>To activate your account, you will need to choose a password, and enter it twice below to make sure ';
        print "you haven't made a mistake in typing.</i><br>";
        print '<table><tr><td>Password:</td><td><input type="password" name="newpassword1" size="20" maxlength="20"></td></tr>';
        print '<tr><td>Type the same password again:</td>'
                                          .'<td><input type="password" name="newpassword2" size="20" maxlength="20"></td></tr></table>';
        print '<p><i>Please enter your e-mail address. This is optional, but you may miss important information about the class if ';
        print "you don't give an address. E-mail is also required in order to reset a forgotten password. ";
        print "Nobody outside of the class will know this address.</i><br>";
        print '<input type="text" name="email" size="50" maxlength="50"><br>';
        print '<input type="checkbox" name="emailpublic" checked value="public"> Leave this box checked if you want other students in ';
        print 'the class to have access to this e-mail address.<br>';
      }
      print '  <input type="submit" value="Log in.">';
      print "</form>\n";
      print "If you're not ".$tree->get_real_name($username,"firstlast")
        .', <a href="'.Url::link(DELETE=>'username').'">click here</a>.<p>';
      print "You must have cookies enabled in your browser in order to log in.<p>\n";
      if ($state eq 'normal') {
        print "<p><i>Forgot your password?</i><br>\n";
        print "If you've forgotten your password, enter your student ID number and click on this button. Information will be e-mailed to you about ";
        print "how to set a new password.<br>\n";
        print '<form method="POST" action="'.Url::link(DELETE=>'login',REPLACE=>'what',REPLACE_WITH=>'emailpassword',
                                                                       DELETE=>'login').'">';
        print '  Student ID: <input type="hidden" name="username" value="'.$username.'">';
        print '  <input type="text" name="id" size="10"> ';
        print '  <input type="submit" value="Send e-mail.">';
        print "</form>\n";
      }
    }
  }
}

#----------------------------------------------------------------
# XML::Parser handlers
#----------------------------------------------------------------

#----------------------------------------------------------------
# char handler
#----------------------------------------------------------------

sub char_handler {
  my ($p,$data) = @_;

  my $debug = 0;


  $data =~ s/^\s*//; # I get 6 extra invisible chars on the front, presumably because
                                          # of something to do with UTF-8. Strip them off.
  if ($data eq "") {return;}
  
  my $depth = Toc::par_depth(); #1=chapter specified, 2=prob specified, 3=find specified
  if ($depth==0) {return;}
  if (!Toc::matches_url()) {return;}

  # Sanity_check checks for length, changes < to &lt; .
  # Super_and_sub changes stuff like ^{2} to <sup>2</sup> .
    $data = SpotterHTMLUtil::super_and_sub(SpotterHTMLUtil::sanity_check(TEXT=>$data,FORBID_LESS_THAN=>1));
  
  if ($debug) {
    print "starting char handler, depth=$depth<br/>\n";
    print "<ul>".Tag_stack::debug_dump()."</ul><br/>\n";
  }

  if ($depth<=2) {
      if (Tag_stack::top()->tag() eq "find") {
        $data =~ s/f\{([^\}]*)\}//g;
        Tag_stack::top()->append_to_character_data(" ".$data);
      }
  }
  if ($depth==3 && !Problem_stack::is_empty()) {
      if (Tag_stack::top()->tag() eq "find") {
        my ($u,$v) = ("\n<br/><img src=\"","\"/><br/>\n");
        $data =~ s/f\{([^\}]*)\}/$u$1$v/g;
        Tag_stack::top()->append_to_character_data(" ".$data);
        Problem_stack::top()->description(Tag_stack::top()->character_data());
      }
      if (Tag_stack::top()->tag() eq "var" && (Url::par("find") eq Problem_stack::top()->id())) {
        my $sym = Tag_stack::top()->att("sym");
        my $vbl = Problem_stack::top()->get_vbl($sym);
        $vbl->description($data);
      }
      if (Tag_stack::top()->tag() eq "ans" && (Url::par("find") eq Problem_stack::top()->id())) {
        my $n = Problem_stack::top()->n_ans();
        my $ans = Problem_stack::top()->get_ans($n);
        $ans->response($ans->response()." ".$data);
      }
  }
  if ($debug) {
    print "done with char handler<br/>\n";
    print "<ul>".Tag_stack::debug_dump()."</ul><br/>\n";
  }
}
#----------------------------------------------------------------
# start handler
#----------------------------------------------------------------
sub start_handler {
  my $p = shift;
  my $el = shift;
  $parsed_any = 1;
  my $debug = 0;
  
  if ($debug) {print "entering start handler <br/>"}
  my $depth = Toc::par_depth(); #1=chapter specified, 2=prob specified, 3=find specified
  my $tag = Tag->new($el,\@_);
  Tag_stack::push($tag);
  if ($debug) {
    print "start $el,".Tag_stack::top()->tag()."<br/>\n";
    print "<ul>".Tag_stack::debug_dump()."</ul><br/>\n";
  }
  if ($el eq "spotter") {
    my $title_of_whole_file = $tag->att("title"); # Defaults to "Contents"
    print "<p><a href=\"".Url::link(DELETE_ALL=>1)."\"><b>$title_of_whole_file</b></a></p>";
  }
  if ($el eq "log_file") {
    Log_file::set_name($basic_file_name,$tag->att("ext"));
  }
  if ($el eq "toc_level") {
    Toc::add_level($tag->att("level"),$tag->att("type"));
  }
  if ($el eq "toc") {
    my ($type,$title,$num) = ($tag->att("type"),$tag->att("title"),$tag->att("num"));
    Toc::start_toc_tag($type,$num,$title);
    if (Toc::matches_url() && (Url::par_is($type,$num) || Toc::toc_depth_specified()>=Toc::depth($type)-1)) {
      my $d = "problem|find|login";
      for (my $i=Toc::current_toc_level()+1; $i<=Toc::toc_depth_specified(); $i++) {
        $d = $d . "|" . $Toc::levels[$i];
      }
      for (my $i=1; $i<= Toc::current_toc_level(); $i++) {print "&mdash;"}
      print "<a href=\"".Url::link(REPLACE=>$type,REPLACE_WITH=>$num,DELETE=>("(".$d.")"))
                                      ."\"><b>". $type." ".$num.": ".$title."</b></a><br>\n";
    }
  }
  if ($el eq "num") {
    my ($id,$label) = ($tag->att("id"),$tag->att("label"));
    if (exists($Num::label{$id})) {print "Fatal error: duplicate id=".$id." in &lt;num> tag."; die}
    $Num::label{$id} = $label;
  }
  if ($el eq "problem") {
    my ($type,$id) = ($tag->att("type"),$tag->att("id"));
    my $matches = Toc::matches_url();
    if (Toc::toc_fully_specified() && $matches) {
      my $l = $Num::label{$id};
      $current_problem = $l;
      if (Url::par_is("problem",$l)) {
        my $problem = Problem->new();
        Problem_stack::push($problem);
      }
      if (!Url::par_set("problem") || Url::par_is("problem",$l)) {
        if (!$printed_any_problems_yet) {print "<p/>\n"}
        $printed_any_problems_yet = 1;
        print "<a href=\"".Url::link(REPLACE=>"problem",REPLACE_WITH=>$l,DELETE=>"find")
                                      ."\">problem ".$l . "</a><br/>\n";
      }
    }
  }
  if (!Problem_stack::is_empty() && $depth==3) {
    if ($el eq "var" && Url::par_is("find",Problem_stack::top()->id())) {
      my ($sym,$type,$min,$max,$min_imag,$max_imag,$units) 
       = ($tag->att("sym"),$tag->att("type"),
          $tag->att("min"),$tag->att("max"),
          $tag->att("min_imag"),$tag->att("max_imag"),$tag->att("units"));
       my $vbl = Vbl->new($sym);
       $vbl->type($type);
       $vbl->min($min);
       $vbl->max($max);
       $vbl->min_imag($min_imag);
       $vbl->max_imag($max_imag);
       $vbl->units($units);
       Problem_stack::top()->add_vbl($vbl);
    } # end if var
    if ($el eq "find") {
      if (Toc::toc_fully_specified() && Toc::matches_url()) {
        my ($id) = ($tag->att("id"));
        if (Url::par_is("find",$id)) {
          Problem_stack::dup();
          Problem_stack::top()->id($id);
        }
      }
    } # end if find
    if ($el eq "ans" && Url::par_is("find",Problem_stack::top()->id())) {
       my ($e,$filter,$tol,$tol_type) 
            = ($tag->att("e"),$tag->att("filter"),$tag->att("tol"),$tag->att("tol_type"));
       my $ans = Ans->new($e,$filter,$tol,$tol_type);
       Problem_stack::top()->add_ans($ans);
    } # end if ans
    if ($el eq "options") {
      Problem_stack::top()->options_stack_dup();
      Problem_stack::top()->options_stack_top()->unit_list($tag->att("unit_list"));
      Problem_stack::top()->options_stack_top()->units_allowed(lc($tag->att("units")) eq 'yes');
        #...should make a special-purpose routine, Options->overwrite(UNIT_LIST=>..., )
    } # end if options
  } # end if !Problem_stack::is_empty() && $depth==3
  if ($debug) {print "leaving start handler<br/>"}
}
#----------------------------------------------------------------
# end handler
#----------------------------------------------------------------
sub end_handler {
  my ($p,$el) = @_;
  my $debug = 0;

  if ($debug) {print "entering end handler"}

  if ($debug) {
    print "end $el,".Tag_stack::top()->tag()."<br/>\n";
    print "<ul>".Tag_stack::top()->debug_dump()."</ul><br/>\n";
    print "<ul>".Tag_stack::debug_dump()."</ul><br/>\n";
  }

  my $depth = Toc::par_depth(); #1=chapter specified, 2=prob specified, 3=find specified
  if ($el eq "toc") {
    Toc::end_toc_tag();
  }
  if (Toc::matches_url()) {
    if ($depth==1) {
      if ($el eq "find") {
        my $max_length = 60;
        my $data = Tag_stack::top()->character_data();
        my $display = $data;
        if (length($data)>$max_length) {
          $display = substr($data,0,$max_length-3); # -3 to allow for ... at end
          # close off any tag that we chopped off in the middle
          while (((length $display)<(length $data)) && $display =~ m/\<[^\>]*$/) {
            $display = substr($data,0,(1+length $display));
          }
          # chopped off in the middle of an &lt; or something (bug: ignorant of quotes)
          while (((length $display)<(length $data)) && $display =~ m/\&\w*$/) {
            $display = substr($data,0,(1+length $display));
          }
          # add any trailing tags
          #  bug: misunderstands quoted angle brackets
          #  bug: leaves in trailing open-tag-close-tag pairs
          my $tail = substr($data,(length $display));
          while ($tail =~ m/(\<[^\>]+\>)/g) {
            $display = $display . $1;
          }
          $display = $display . "...";
        }
        #print "&mdash;".$display."<br/>\n";
        print "&mdash;<a href=\""
          .Url::link(REPLACE=>"find",REPLACE_WITH=>Tag_stack::top()->att("id"))
          ."&problem=$current_problem\">".$display."</a><br/>\n";
      } # end if find
    } # end if depth==1
    if ($el eq "options" && !Problem_stack::is_empty() && $depth==3) {
      Problem_stack::top()->options_stack_pop();
    }
    if ($el eq "problem" && !Problem_stack::is_empty()) {
      Problem_stack::pop();
    }
    if ($depth==2 && !Problem_stack::is_empty()) {
      if (Tag_stack::top()->tag() eq "find") {
        print "&mdash;<a href=\""
         .Url::link(REPLACE=>"find",REPLACE_WITH=>Tag_stack::top()->att("id"))
         ."\">".Tag_stack::top()->character_data()."</a><br/>";
        #print "<p>".Tag_stack::top()->character_data()."</p>";
      }
    }
    if ($depth==3 && Problem_stack::depth()>=1) {
      if (Tag_stack::top()->tag() eq "find") {

        my $debug = 0;

        if ($debug) {print "entering end handler for find"; print Problem_stack::debug_dump()."<br/>"}

        my $p = Problem_stack::top();

        my $raw_ans = $SpotterHTMLUtil::cgi->param("answer");

         # The following has to happen /before/ the sanity check, since the sanity check might
         # filter out illegal characters, like =. Therefore, we don't want to print out
         # any portion of the string, or it would be risky in terms of security.
         if (index($raw_ans,"=")>=0) {
           $raw_ans = substr($raw_ans,index($raw_ans,"=")+1);
           print "<p>You don't need to type in an equation, just an expression. Everything "
                   ."to the left of the equals sign has been disregarded.</p>";
         }

        my ($ans,$insane) = SpotterHTMLUtil::sanity_check(TEXT=>$raw_ans);
        if ($insane) {print "$insane";}

        # Make sure the raw version doesn't get used inadvertently later:
        undef $raw_ans;

        my $unit_list = "";
        my $units_allowed = 0;
        if ($p->options_stack_not_empty()) {
          $unit_list = $p->options_stack_top()->unit_list();
          $units_allowed = $p->options_stack_top()->units_allowed(); # gets modified below if it's a numerical answer with a menu of units
        }
        # The following is only for numerical answers that have units. If they entered 37 and selected units of meters
        # from the menu, what we do right here is parse the 37.
        if ($unit_list && $ans ne "") {
          $units_allowed = 1; # only for use by answer_response(), which gets the whole thing, like "37 m"
          my $unit_selection = $SpotterHTMLUtil::cgi->param("unit_list");
          my ($unit_selection,$insane) = SpotterHTMLUtil::sanity_check(TEXT=>$unit_selection);
          my $e = Expression->new(EXPR=>$ans,UNITS_ALLOWED=>0); # disallow units here, because we're only parsing the number
          my $m = Measurement::promote_to_measurement($e->evaluate());
          if ($e->has_errors()) {
            print $e->format_errors();
            $ans = "";
          }
          if ($ans && !($m->reduces_to_unitless(\%Spotter::standard_units))) {
            print "<p>For this problem, don't type in the units. Use the pop-up menu on the right.</p>\n";
            $ans = "";
          }
          if ($ans ne "") {
            $m = $m * Measurement->new(1,Units::parse_units(TEXT=>$unit_selection));
            $ans = "$m";
          }
        }

        print "<p>".$p->description()."</p>\n";
        my @vbl_list = $p->vbl_list();
        if (@vbl_list) {
          print "<p>Variables:\n";
          print "<table>\n";
          foreach my $sym(@vbl_list) {
            my $vbl = $p->get_vbl($sym);
            print "<tr><td><ul>".$sym."</ul></td><td></td><td></td><td></td><td></td><td>".$vbl->description()."</td></tr>\n";
          }
          print "</table>\n";
        }

        my ($response,$student_answer_is_correct);
        # Throttle file:
          my $query = $ENV{'QUERY_STRING'};
          my $ip = $ENV{REMOTE_ADDR};
          my $date_string = current_date_string_no_time();
          my $throttle_dir = "spotter/throttle";
          if (! -e $throttle_dir) {mkdir($throttle_dir)}
          # Grant exemptions to throttling for certain answer files, e.g., demo files.
            my $exempt = 0;
            if (open(FILE,"<$throttle_dir/exempt_files")) {
            while (my $line=<FILE>) {
              chomp $line;
              if ($line eq $xmlfile) {$exempt=1}
            }  
            close(FILE);
          }
          # Forbid anonymous use from certain addresses.
            my $anon_forbidden = 0;
            if (open(FILE,"<$throttle_dir/forbid_anonymous_use_at")) {
            while (my $line=<FILE>) {
              chomp $line;
              my @a = split /\./,$ip;
              my @b = split /\./,$line;
              my $match = 1;
              for (my $i=0; $i<=$#b; $i++) {
                if ($a[$i] != $b[$i]) {$match=0}
              }
              if ($match && $#b>=0) {$anon_forbidden=1}
            }  
            close(FILE);
          }
          my $when = time;
          my $who = '';
          if ($login->logged_in()) {$who = $login->username()}
            my $forbidden_because_anon = ($anon_forbidden && $who eq '');
          my $throttle_file_name = "$throttle_dir/log_$date_string";
          my $query_sha1 = Digest::SHA1::sha1_base64($p->description()); # was originally based on query, but that's too easy to spoof
          my $throttle_ok = 1;
          my $when_over;
          my %n_within = ();
          my %max_within = (
            20 => 1,  # no more than 1 in 10 seconds
            20 => 3,  # no more than 3 in 20 seconds
            180 => 5, # no more than 5 in 3 minutes
            600 => 8, # no more than 8 in 10 minutes
            3600 => 15, # no more than 15 in 1 hour
          );
          my $max_per_day = 30; # each throttle file is 1 day
          my $longest_interval_violated = 0;
            if ($ans ne "") {
            if (open(FILE,"<$throttle_file_name")) {
              my @times = ();
              while (my $line = <FILE>) {
                chomp $line;
                my ($ip_was,$who_was,$when_was,$query_sha1_was) = split /,/,$line;
                if ((($ip_was eq $ip && ($who_was eq '' || $who eq '')) || $who_was eq $who) && $query_sha1 eq $query_sha1_was) {
                  push @times,$when_was;
                }
              }
              close(FILE);
              my $n_times = ($#times)+1;
              if ($n_times>=$max_per_day) {$throttle_ok = 0}
              if ($throttle_ok) {
                my @intervals = keys %max_within;
                foreach my $interval(@intervals) {
                  my @within_this_interval = ();
                  $n_within{$interval} = 0;
                  foreach my $time(@times) {
                    if ($time+$interval>$when) {
                      ++$n_within{$interval};
                      push @within_this_interval,$time;
                    }
                  }
                  if ($n_within{$interval}>=$max_within{$interval}) {
                    $throttle_ok = 0;
                    if ($interval>$longest_interval_violated) {
                      $longest_interval_violated = $interval;
                      @within_this_interval = sort {$b <=> $a} @within_this_interval; # reverse numerical order
                      $when_over = $within_this_interval[$max_within{$interval}-1]+$interval-$when;
                    }
                  }
                }
              }
            }
          }
        my $reason_forbidden = '';
        my $exempt_message = '';
        if (!$throttle_ok) {
          $reason_forbidden=
           "Too many answers have been entered in a short time period on this computer and/or this account."
            . " To discourage random guessing, longer and longer wait times are required if you"
            . " keep on entering answers over and over."
            . " If you're having trouble doing this problem, maybe you should get help from your instructor!"
            . " You have entered more than ".$max_within{$longest_interval_violated}." answers within "
            . $longest_interval_violated." seconds."
            . " This waiting period will expire in ".$when_over." seconds.";
          $exempt_message = "(This answer file is exempt from waiting time requirements, but "
              . " you have entered more than ".$max_within{$longest_interval_violated}." answers within "
              . $longest_interval_violated." seconds.)";
          my $add_on = '';
          if ($who eq '') {
            $add_on = 
              " (Since you're using Spotter anonymously, you may get this message even if it's the first"
              . " time you've attempted the problem. This is because Spotter considers all anonymous"
              . " users to be the same person.)"
            }
          $reason_forbidden = "$reason_forbidden$add_on";
          $exempt_message = "$exempt_message$add_on";
          $exempt_message =~ s/\)\s+\(/ /g; # don't put two parenthetical statements in a row; combine them instead
        }
        if ($forbidden_because_anon) {
          $throttle_ok = 0;
          $reason_forbidden="Anonymous access is not allowed from your location. Please log in.";
          $exempt_message="(Anonymous access is not normally allowed from your location, "
                           ."but this answer file is exempt from that restriction.)";
        }
        if (!$throttle_ok) {
          print '<table width="400"><tr><td>';
          if (!$exempt) {
            print $reason_forbidden;
          }
          else {
            print $exempt_message;
            $throttle_ok = 1;
          }
          print"</td></tr></table>\n";
        }
        if ($ans ne "" && $throttle_ok) {
          # The user gave an answer, and clicked the Check button.
          print "<p>Your answer was $ans .</p>";
          # The units_allowed feature doesn't quite work right, so if units are not allowed, strip out all spaces
          # from the answer. This has the effect of making sure 2 g is interpreted as 2*g, not 2 grams. Same deal for 2 m.
					# Don't filter out all blanks, because they could be significant in, e.g., 3 10^3.
          my $ans2 = $ans;
          if (!$units_allowed) {
            $ans2 =~ s/\s+g/g/g;
            $ans2 =~ s/\s+m/m/g;
          }
          ($response,$student_answer_is_correct) =  AnswerResponse::answer_response($p,$ans2,$units_allowed);
          print $response;
        }
        print "\n".$SpotterHTMLUtil::cgi->startform."\n";
        print "<p>";
        my $default = "";
        my $confirm_recorded = 0;
        if ($ans ne "" && $throttle_ok) {
          if ($student_answer_is_correct) {
            print "Your correct answer was:<br/>";
          }
          else {
            print "Try again:<br/>";
          }
          if ($login->logged_in()) {
            record_work(LOGIN=>$login,FILE_TREE=>$tree,ANSWER=>$ans,IS_CORRECT=>$student_answer_is_correct,
                     RESPONSE=>$response,DESCRIPTION=>$p->description(),QUERY=>$query);
            $confirm_recorded = $student_answer_is_correct;
          }
          $default=$ans;
          # Throttle file:
            if (open(FILE,">>$throttle_file_name")) {
              print FILE "$ip,$who,$when,$query_sha1\n";
              close(FILE);
            }
        } 
        else {
          print "Answer:<br/>"
        }
        print $SpotterHTMLUtil::cgi->textfield(-name=>"answer",-default=>$default,-size=>"60");
                # For some reason, the -default=> doesn't seem to have the effect described in
                # the documentation.
        if ($unit_list) {
          my @unit_list = split (/\,/ , $unit_list);
          print $SpotterHTMLUtil::cgi->popup_menu(-name=>'unit_list',-values=>\@unit_list);
        }
        print $SpotterHTMLUtil::cgi->submit(-value=>"Check")."\n";
        print "</p>\n";
        print "\n".$SpotterHTMLUtil::cgi->endform."\n";

        if ($confirm_recorded) {
            print "<p>This correct answer was recorded under your name, "
                .$tree->get_real_name($login->username(),"firstlast").".</p>\n";
        }
        
        print SpotterText::how_to_enter_answers();
        
        Problem_stack::pop();
        if ($debug) {print "leaving end handler for find"}
      }
    }
  }
  Tag_stack::pop();
  if ($debug) {print "leaving end handler"}
}

sub record_work {
  my %args = (
    ANSWER=>'',
    DESCRIPTION=>'',
    LOGIN=>'',
    FILE_TREE=>'',
    IS_CORRECT=>'',
    RESPONSE=>'',
    QUERY=>'',
    @_,
  );
  my $record_response = "";
  if (!$args{IS_CORRECT}) {
    $record_response = $args{RESPONSE};
    $record_response =~ s/\n/ /g;
    $record_response =~ s/^\<p\>//;
    if (length($record_response)>64) {$record_response=substr($record_response,0,64)}
  }
  my $description = $args{DESCRIPTION};
  $description =~ s/\n / /g;
  $description =~ s/^\s+//;
  if (length($description)>64) {$description=substr($description,0,64)}
  my $answer = $args{ANSWER};
  $answer =~ s/\n/ /g;
  my $login = $args{LOGIN};
  my $tree = $args{FILE_TREE};
  my $is_correct = 0;
  if ($args{IS_CORRECT}) {$is_correct=1}
  my $file = $tree->student_work_file_name($login->username());
  my $query = $args{QUERY};
  $query =~ s/username=[^\&]*\&?//;
  $query =~ s/class=[^\&]*\&?//;
  $query =~ s/what=check\&?//;
  open(FILE,">>$file") or return "error opening work file";
  my $nlines = 6;
  if ($record_response) {
    $nlines = $nlines+1;
  }
  my $ip = $ENV{REMOTE_ADDR};
  print FILE "answer,$nlines\n";
  print FILE "  query:       $query\n";
  print FILE "  description: $description\n";
  print FILE "  answer:      $answer\n";
  print FILE "  correct:     $is_correct\n";
  print FILE "  date:        ".current_date_string()."\n";
  print FILE "  ip:          $ip\n";
  if ($record_response) {
    print FILE "  response:    $record_response\n";
  }
  close(FILE);
  return 0;
}



#----------------------------------------------------------------
# Options class
#
# When we create a new Problem object, we set its options
# using Options->new(), using all the defaults. When we hit
# an <options> tag, we push onto that problem's options stack.
# Then any new Ans objects get created with this Problem's
# options.
# When we hit the <options/> tag, we pop the options stack.
#----------------------------------------------------------------
package Options;
sub new {
  my $class = shift;
  my %args = (
    UNIT_LIST => "",
    @_
  );
  my $self = {};
  bless($self,$class);
  if ($args{UNIT_LIST}) {$self->unit_list($args{UNIT_LIST})}
  return $self;
}

sub unit_list {
  my $self = shift;
  if (@_) {$self->{UNIT_LIST} = shift;}
  if (exists($self->{UNIT_LIST})) {return $self->{UNIT_LIST}}
  return "";
}

sub units_allowed {
  my $self = shift;
  if (@_) {$self->{UNITS_ALLOWED} = shift;}
  if (exists($self->{UNITS_ALLOWED})) {return $self->{UNITS_ALLOWED}}
  return "";
}

sub debug_dump {
  my $self = shift;
  return "Options::debug_dump, unit_list=".$self->unit_list()."<br>\n";
}

#----------------------------------------------------------------
# Ans class
#----------------------------------------------------------------
package Ans;
sub new {
  my $class = shift;
  my ($e,$filter,$tol,$tol_type) = (@_);
  my $self = {};
  bless($self,$class);
  $self->e($e);
  $self->filter($filter);
  $self->tol($tol);
  $self->tol_type($tol_type);
  $self->response("");
  $self->options(Problem_stack::top()->options_stack_top());
  # Note that all the defaults get filled in by the XML parser.
  return $self;
}

sub is_correct {
  my $self = shift;
  return $self->response() eq "";
}

sub debug_dump {
  my $self = shift;
  return "e=".$self->e().", filter=".$self->filter().", tol="
          .$self->tol().", tol_type=".$self->tol_type().", response="
          .$self->response().", is_correct=".$self->is_correct();
}

sub response {
  my $self = shift;
  if (@_) {$self->{RESPONSE} = shift;}
  return $self->{RESPONSE};
}

sub e {
  my $self = shift;
  if (@_) {$self->{E} = shift;}
  return $self->{E};
}

sub filter {
  my $self = shift;
  if (@_) {$self->{FILTER} = shift;}
  return $self->{FILTER};
}

sub tol {
  my $self = shift;
  if (@_) {$self->{TOL} = shift;}
  return $self->{TOL};
}

sub tol_type {
  my $self = shift;
  if (@_) {$self->{TOL_TYPE} = shift;}
  return $self->{TOL_TYPE};
}

sub options {
  my $self = shift;
  if (@_) {$self->{OPTIONS} = shift;}
  return $self->{OPTIONS};
}



#----------------------------------------------------------------
# Vbl class
#----------------------------------------------------------------
package Vbl;
sub new {
  my $class = shift;
  my $sym = shift;
  my $self = {};
  bless($self,$class);
  $self->{SYM} = $sym;
  # Note that all the defaults get filled in by the XML parser.
  return $self;
}

sub debug_print {
  my $self = shift;
  print $self->debug_dump();
}

sub debug_dump {
  my $self = shift;
  my $result = "<p>";
  $result = $result . $self->sym().", ".$self->description().", ".$self->type().", ";
  $result = $result . "real(".$self->min().",".$self->max()."), ";
  $result = $result . "imag(".$self->min_imag().",".$self->max_imag().")</p>\n";
  return $result;
}

sub sym {
  my $self = shift;
  if (@_) {$self->{SYM} = shift;}
  return $self->{SYM};
}

sub description {
  my $self = shift;
  if (@_) {$self->{DESCRIPTION} = shift;}
  return $self->{DESCRIPTION};
}

sub type {
  my $self = shift;
  if (@_) {$self->{TYPE} = shift;}
  return $self->{TYPE};
}

sub units {
  my $self = shift;
  if (@_) {$self->{UNITS} = shift;}
  return $self->{UNITS};
}

sub min {
  my $self = shift;
  if (@_) {$self->{MIN} = shift;}
  return $self->{MIN};
}

sub max {
  my $self = shift;
  if (@_) {$self->{MAX} = shift;}
  return $self->{MAX};
}

sub min_imag {
  my $self = shift;
  if (@_) {$self->{MIN_IMAG} = shift;}
  return $self->{MIN_IMAG};
}

sub max_imag {
  my $self = shift;
  if (@_) {$self->{MAX_IMAG} = shift;}
  return $self->{MAX_IMAG};
}

sub parsed_units {
  my $self = shift;
  if (@_) {$self->{PARSED_UNITS} = shift;}
  return $self->{PARSED_UNITS};
}


#----------------------------------------------------------------
# Problem_stack class
#----------------------------------------------------------------
package Problem_stack;
my @problem_stack = []; #Each element is a reference to a Find.

sub push {
  my $problem = shift;
  push @problem_stack,$problem;
}

sub dup {
  push @problem_stack,top();
}

sub pop {
  pop @problem_stack;
}

sub top {
  return $problem_stack[$#problem_stack];
}

sub is_empty {
  return $#problem_stack<0;
}

sub depth {
  return $#problem_stack;
}

sub debug_dump {
  my $x = "problem stack: ";
  for (my $i=0; $i<=$#problem_stack; $i++) {
    $x = $x . "$i ";
  }
  return $x;
}

#----------------------------------------------------------------
# Problem class
#----------------------------------------------------------------
package Problem;

sub new {
  my $class = shift;
  my $id = shift;
  my $self = {};
  bless($self,$class);
  $self->{ID} = $id;
  my %empty = ();
  $self->{VBLS} = \%empty;
  my @empty = ();
  $self->{ORDERED_VBLS} = \@empty;
  my @empty2 = ();
  $self->{ANSWERS} = \@empty2; # does [] work?
  $self->{OPTIONS_STACK} = [Options->new()];
  return $self;
}

sub vbl_hash {
  my $self = shift;
  my $h = $self->{VBLS};
  return %$h;
}

sub vbl_list {
  my $self = shift;
  my $r = $self->{ORDERED_VBLS};
  return @$r;
}

sub add_vbl {
  my $self = shift;
  my $vbl = shift;
  my $vbls = $self->{VBLS};
  $vbls->{$vbl->sym()}=$vbl;
  my $r = $self->{ORDERED_VBLS};
  push @$r,$vbl->sym();
}

sub add_ans {
  my $self = shift;
  my $ans = shift;
  my $answers = $self->{ANSWERS};
  push @$answers,$ans;
}

sub n_ans {
  my $self = shift;
  my $answers = $self->{ANSWERS};
  return $#$answers;
}

sub get_ans {
  my $self = shift;
  my $i = shift;
  if ($i>$self->n_ans()) {return "";}
  my $answers = $self->{ANSWERS};
  return $answers->[$i];
}

sub get_vbl {
  my $self = shift;
  my $sym = shift;
  my $vbls = $self->{VBLS};
  return $vbls->{$sym};
}

sub id {
  my $self = shift;
  if (@_) {$self->{ID} = shift;}
  return $self->{ID};
}

sub description {
  my $self = shift;
  if (@_) { $self->{DESCRIPTION} = shift;}
  return $self->{DESCRIPTION};
}

sub options_stack_not_empty {
  my $self = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  return ($#stack>=0);
}

sub options_stack_top {
  my $self = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  if ($#stack>=0) {return $stack[$#stack]}
  return "";
}

sub options_stack_push {
  my $self = shift;
  my $o = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  push @stack , $o;
  $self->{OPTIONS_STACK} = \@stack;
}

sub options_stack_pop {
  my $self = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  pop @stack;
  $self->{OPTIONS_STACK} = \@stack;
}

sub options_stack_dup {
  my $self = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  my $o_ref = $stack[$#stack];
  my %o = %$o_ref;
  my %o_clone = %o;
  bless(\%o_clone,"Options");
  push @stack , \%o_clone;
  $self->{OPTIONS_STACK} = \@stack;
}

sub debug_dump {
  my $self = shift;
  my $stack_ref =  $self->{OPTIONS_STACK};
  my @stack = @$stack_ref;
  my $result = "";
  $result = $result. "Problem::debug_dump, options stack depth=".$#stack."<br>\n";
  for (my $i=0; $i<=$#stack; $i++) {
    $result = $result . $i. ": ".$stack[$i]->debug_dump();
  }
  return $result;
}
#----------------------------------------------------------------
# Tag_stack class
#----------------------------------------------------------------
package Tag_stack;

my @tag_stack = []; #Each element is a reference to a Tag.

sub debug_print {
  print "<p>Tag_stack::debug_print...";
  for (my $i=0; $i<=$#tag_stack; $i++) {
    my $tag = $tag_stack[$i];
    $tag->debug_print();
  }
  print "...done</p>\n";
}

sub push {
  my $tag = shift;
  push @tag_stack,$tag;
}

sub pop {
  pop @tag_stack;
}

sub top {
  return $tag_stack[$#tag_stack];
}

sub contains {
  my $search_tag = shift;
  for (my $i=0; $i<=$#tag_stack; $i++) {
    my $tag = $tag_stack[$i];
    if ($tag->tag() eq $search_tag) {return 1;}
  }
  return 0;
}

sub debug_dump {
  my $msg = "tag stack: ";
  for (my $i=0; $i<=$#tag_stack; $i++) {
    my $tag = $tag_stack[$i];
    $msg = $msg .$tag->tag()." ";
  }
  return $msg;
}

#----------------------------------------------------------------
# Tag class
#----------------------------------------------------------------
package Tag;

# The object is a hash where the keys are names of attributes.
# The name of the tag itself is stored in {$tag_key}

my $tag_key = "!"; # Not a legal XML attribute.

# Create a new tag object.
sub new {
  my $class = shift;
  my ($el,$att_ref) = @_;
  my @atts = @$att_ref;
  my $self = {};
  bless($self,$class);
  $self->{$tag_key} = $el;
  for (my $i=0; $i<$#atts; $i=$i+2) {
    $self->add($atts[$i],$atts[$i+1]);
  }
  $self->character_data("");
  return $self;
}

# Read the name of the tag.
sub tag {
  my $self = shift;
  return $self->{$tag_key};
}

# Character data from inside the tag
sub character_data {
  my $self = shift;
  if (@_) {
    $self->{CHARACTER_DATA} = shift;
  }
  return $self->{CHARACTER_DATA};
}

# Character data from inside the tag
sub append_to_character_data {
  my $self = shift;
  my $c = shift;
  $self->character_data($self->character_data().$c);
}

# Write/overwrite attribute list.
sub add {
  my $self = shift;
  my ($attribute,$value) = @_;
  $self->{$attribute} = $value;
}

# Read attribute list.
# Second argument is default -- not yet implemented.
sub att {
  my $self = shift;
  my ($name) = @_;
  if (exists($$self{$name})) {return $$self{$name}}
  return "";
}


sub debug_print {
  my $self = shift;
  print $self->debug_dump();
}

sub debug_dump {
  my $self = shift;
  my %h = %$self;
  my @k = keys(%h);
  my $msg = "Tag::debug_print ";
  $msg = $msg . "ref self=".ref($self)." ";
  $msg = $msg .  "keys=".$#k." ";
  #$msg = $msg .  "self tag key=".$self->{$tag_key}." ";
  $msg = $msg .  "tag=".$self->tag();
  foreach my $att(keys(%h)) {
    $msg = $msg .  " ".$att."=".$h{$att};
  }
  $msg = $msg .  "\n";
}

#----------------------------------------------------------------
# Url class
#----------------------------------------------------------------
package Url;

our %params = ();

# Construct URL that links back to me. The DELETE
# arg tells which params to delete, written like "a"
# to delete a=xxx only, or "(a|b)" to delete more than one.
# If DELETE_ALL is used, then NOT_DELETE is the list
# not to delete, in the same style. Don't use DELETE and
# DELETE_ALL together.
sub link {
  my %args = (
    DELETE_ALL => 0,
    RELATIVE=>1,
    DELETE=>'',
    NOT_DELETE => "(what|debug|file|class)",
    REPLACE => "",
    REPLACE_WITH => "",
    REPLACE2 => "",
    REPLACE_WITH2 => "",
    REPLACE3 => "",
    REPLACE_WITH3 => "",
    @_,
  );
    
  my $cgi = new CGI;
  my $this_script = $cgi->url(-relative=>$args{RELATIVE});
  
  # Workaround for a bug in CGI.pm: If GET and POST are both used, then
  # query_string only gives the javascript form stuff, not the URL stuff.
  #my $q = $cgi->query_string; # doesn't work
  my $q = "";
  my @u = $cgi->url_param;
  foreach my $a(@u) {
    if ($a ne "") {
      $q = $q . $a . "=" . $cgi->url_param($a) . "&"; # Extra & on the end gets cleaned up later
    }
  }

  # --- DELETE
  for (my $i=1; $i<=2; $i++) {
    my $d;
    if ($i==1) {$d = $args{DELETE}}
    if ($d ne "") {
      $q =~ s/$d=[^=\&]*//g;
    }
  }

  # --- DELETE_ALL
  if ($args{DELETE_ALL}) {
    my $nd = $args{NOT_DELETE};
    if ($nd eq "") {
      $q =~ s/[^=\&]+=[^=\&]*//g;
    }
    else {
      my $new = "";
      while ($q =~ m/($nd=[^=\&]*)/g) {
        $new = $new . $1 . "&";
      }
      $q = $new;
    }
  }
  
  # --- REPLACE
  my $r = $args{REPLACE};
  my $rw = $args{REPLACE_WITH};
  if ($r ne "") {
    if (!($q =~ m/$r/)) {$q=$q."&$r=$rw";}
    $q =~ s/$r=[^=\&]*/$r=$rw/;
  }
  my $r = $args{REPLACE2};
  my $rw = $args{REPLACE_WITH2};
  if ($r ne "") {
    if (!($q =~ m/$r/)) {$q=$q."&$r=$rw";}
    $q =~ s/$r=[^=\&]*/$r=$rw/;
  }
  my $r = $args{REPLACE3};
  my $rw = $args{REPLACE_WITH3};
  if ($r ne "") {
    if (!($q =~ m/$r/)) {$q=$q."&$r=$rw";}
    $q =~ s/$r=[^=\&]*/$r=$rw/;
  }
  
  # Tidy up.
  $q =~ s/\&\&+/\&/g; # replace && with &
  $q =~ s/\&+$//g; # strip & off the end
  
  return $this_script."?".$q;
}

sub par_is {
  my ($par,$val) = @_;
  return par($par) && $params{$par} eq $val;
}

sub par_set {
  my ($par) = @_;
  return exists($params{$par});
}

sub par {
  my ($par) = @_;
  return $params{$par};
}

sub report_pars {
  print "<p>Debugging output: list of GET method parameters:<br/>\n";
  foreach my $par(keys(%params)) {
    print "$par=".$params{$par}."<br/>\n";
  }
  print "</p>\n";
}

sub decode_pars {

        my $my_query_string = $ENV{'QUERY_STRING'};
        #my $my_query_string = "file=lm&what=check";
                
        my @query_key_pairs = split(/&/, $my_query_string);
        
        if (! @query_key_pairs) {return 0;}
        
        %params = ();
        foreach my $par (@query_key_pairs) {
          my ($a,$b) = split(/=/, $par);
          my ($aa,$bb,$msg);
          $aa = $a;
          $bb = $b;
          ($aa,$msg) = SpotterHTMLUtil::sanity_check(TEXT=>$a,MAX_LENGTH=>50);
          ($bb,$msg) = SpotterHTMLUtil::sanity_check(TEXT=>$b,MAX_LENGTH=>50);
          $params{$aa} = $bb;
        }
}


#----------------------------------------------------------------
# Num class
#----------------------------------------------------------------
package Num;
our %label = ();

#----------------------------------------------------------------
# Toc class
#----------------------------------------------------------------
package Toc;

our @levels = []; # "chapter", "section", etc.
our @stack_num = [];
our @stack_title = [];

# The following two routines should check for correct nesting, but don't.
# The $type argument is to allow for this.
sub start_toc_tag {
  my ($type,$num,$title) = @_;
  push (@stack_num,$num); 
  push (@stack_title,$title); 
  #print "<p>start_toc_tag $type,$num,$title</p>\n";
}
sub end_toc_tag {
  pop @stack_num; 
  pop @stack_title; 
}

sub depth {
  my $type = shift;
  for (my $i=0; $i<=deepest_toc_level(); $i++) {
    if ($levels[$i] eq $type) {return $i;}
  }
  return -1;
}

sub par_depth {
  if (!toc_fully_specified()) {return 0;}
  if (!Url::par_set("problem")) {return 1;}
  if (!Url::par_set("find")) {return 2;}
  return 3;
}

sub matches_url {
  for (my $i=0; $i<=toc_depth_specified() && $i<=deepest_toc_level(); $i++) {
    my $l = $levels[$i];
    if (Url::par($l) ne $stack_num[$i] && $stack_num[$i] ne "") {
      #print "<p>mismatch, deepest=".deepest_toc_level()." i=$i, l=$l, ".Url::par($l).", =".$stack_num[$i]."=</p>";
      return 0;
    }
  }
  return 1;
}

sub toc_fully_specified {
  return toc_depth_specified()==deepest_toc_level();
}

sub toc_depth_specified {
  for (my $i=0; $i<=deepest_toc_level(); $i++) {
    if (!Url::par_set($levels[$i])) {return $i-1;}
  }
  return deepest_toc_level();
}

sub current_toc_level {
  return $#stack_num;
}

sub deepest_toc_level {
  return $#levels;
}

sub add_level {
  my ($level,$type) = @_;
  $levels[$level] = $type;
}
#----------------------------------------------------------------
# Log_file class
#----------------------------------------------------------------
package Log_file;

my $log_file_name = "log";

sub set_name {
  my $basic_file_name = shift;
  my $ext = shift;
  $ext =~ m/^\.?([a-zA-Z0-9\_]{1,30})$/;
  if (!$1) {$ext="log"} else {$ext = $1;}
  $log_file_name = $basic_file_name . "." . $ext;
  SpotterHTMLUtil::debugging_output("Setting log file to $log_file_name");
}

sub get_name {
  return $log_file_name; 
}

sub write_entry {
  my %args = (
    TEXT => "",
    @_
  );
  my $text = $args{TEXT};
  #print "logging text=$text=<br/>\n";
  
  my ($s,$min,$h,$d,$mo,$y) = (localtime)[0..5];
  if ($y<1900) {$y=$y+1900} # This should work in either Perl 5 or Perl 6.
  my $stuff = sprintf ( "%04d-%02d-%02d %02d:%02d:%02d ", $y, ($mo+1), $d, $h, $min, $s) 
    . $text
          . "\n";
  
  
  my $ok = open(LOG_FILE, ">>$log_file_name");
  if (!$ok) {
    print "logging failed\n";
    return;
  }
  print LOG_FILE $stuff;
  close(LOG_FILE);
 }

